import json
import os.path

from pkg_resources import resource_filename


class Calibration(object):

    def __init__(self, *args, **kwargs):
        """
        Calibration(0.34) - 0.34 pixel per micrometer OR
        Calibration(microscope = "Zeiss", objective = "20x")
        """
        try:
            self.pxPerMicron = float(args[0])
        except ValueError:
            raise ValueError(
                "invalid argument %s: need to give numeric value \
                 for pxPerMicron" % args[0])
        except IndexError:
            microscope_name = kwargs.get("microscope")
            objective = kwargs.get("objective")

            if "calibration_file" in kwargs:
                calibration_file = kwargs.get("calibration_file")
            else:
                calibration_file = resource_filename(
                    "spotlob", "resources/calibrations.json")

            assert os.path.exists(calibration_file)

            caldict = Calibration.read_calibration_file(calibration_file)

            self.pxPerMicron = float(caldict[microscope_name][objective])

    def pixel_to_micron(self, pixels):
        return pixels / self.pxPerMicron

    def squarepixel_to_squaremicron(self, pixels_squared):
        return pixels_squared / self.pxPerMicron**2

    def micron_to_pixel(self, microns):
        return microns * self.pxPerMicron

    def squaremicron_to_squarepixel(self, microns_squared):
        return microns_squared * self.pxPerMicron**2

    @classmethod
    def read_calibration_file(cls, calibration_file="calibrations.json"):
        with open(calibration_file, "rb") as f_:
            caldict = json.load(f_)
        return caldict

    @classmethod
    def microscope_in_file(cls, calibration_file="calibrations.json"):
        caldict = Calibration.read_calibration_file(calibration_file)
        return caldict.keys()

    @classmethod
    def lenses_in_file(cls, microscope, calibration_file="calibrations.json"):
        caldict = Calibration.read_calibration_file(calibration_file)
        return caldict[microscope].keys()

    def calibrate(self, dataframe):
        """
        get all columns that include the suffix _px and _px2 
        and calculate additional columns with the micron / micron^2 values
        """
        for col in dataframe.columns:
            if col.endswith("_px"):
                new_col_name = col[:-3] + "_um"
                dataframe[new_col_name] = self.pixel_to_micron(
                    dataframe[col])
            elif col.endswith("_px2"):
                new_col_name = col[:-4] + "_um2"
                dataframe[new_col_name] = self.squarepixel_to_squaremicron(
                    dataframe[col])
        return dataframe



def parameter_from_spec(spec):
    """This function will create a SpotlobParameter from a specification

    PARAMETERS
    ----------
    spec : tuple(str, object)
        specification for the parameter, must be one of the following
        options:

        +---------------+--------------------------------------+
        | float range   | ("parameter_name", (float_min_value, |
        |               |                     float_max_value, |
        |               |                     float_value))    |
        +---------------+--------------------------------------+
        | integer range | ("parameter_name", (int_value,       |
        |               |                     int_min_value,   |
        |               |                     int_max_value))  |
        +---------------+--------------------------------------+
        | boolean value | ("parameter name", boolean)          |
        +---------------+--------------------------------------+
        | enumeration   | ("parameter name", ["option1",       |
        |               |                     "option2",       |
        |               |                     "option3"])      |
        +---------------+--------------------------------------+

    RETURNS
    -------
    SpotlobParameter
        An instance of a SpotlobParameter subclass: EnumParameter,
        FloatParameter, ... depending on the type of the spec

    """
    try:
        parname, val = spec
        try:
            # split into min,max,value
            minv, maxv, v = val
            if any([type(vi) == float for vi in val]):
                return NumericRangeParameter(parname,
                                             float(v),
                                             float(minv),
                                             float(maxv),
                                             float)
            elif all([type(vi) == int for vi in val]):
                return NumericRangeParameter(parname,
                                             int(v),
                                             int(minv),
                                             int(maxv),
                                             int)
            else:
                raise TypeError

        except TypeError:
            # could not create a slider
            if type(val) == bool:
                return BoolParameter(parname, val)
            elif all([type(s) == str for s in val]):
                return EnumParameter(parname, val[0], val)
    except:
        raise Exception("Invalid parameter specification")


class SpotlobParameter(object):
    def __init__(self, name, value, type_, description=""):
        self.name = name
        self._value = value
        self.type = type_
        self.description = description
        self.preview_enabled = False
        super(SpotlobParameter, self).__init__()

    def __repr__(self):
        return "<SpotlobParameter(%s) %s: %s>"\
            % (self.type, self.name, self.value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_val):
        self._value = new_val

    def __str__(self):
        return "%s: %s" % (self.name, self.value)


class SpotlobParameterSet(object):
    def __init__(self, parameters):
        self.parameters = parameters

    @property
    def names(self):
        return [p.name for p in self.parameters]

    @property
    def values(self):
        return [p.value for p in self.parameters]

    def __getitem__(self, identifier):
        try:
            return self.parameters[identifier]
        except TypeError:
            ind = self.names.index(identifier)
            return self.parameters[ind]

    def to_dict(self):
        return dict([(p.name, p.value) for p in self.parameters])

    def __str__(self):
        lines = ["- %s\n" % p for p in self.parameters]
        return "".join(lines)


class FilepathParameter(SpotlobParameter):
    def __init__(self, name, path):
        super(FilepathParameter, self).__init__(name, path, str, "")


class EnumParameter(SpotlobParameter):
    def __init__(self, name, value, options, description=""):
        self.options = options
        super(EnumParameter, self).__init__(
            name, value, str, description)


class NumericRangeParameter(SpotlobParameter):
    def __init__(self, name, value, minvalue, maxvalue,
                 type_=int, step=1, description=""):
        self.minvalue = minvalue
        self.maxvalue = maxvalue
        self.step = step
        super(NumericRangeParameter, self).__init__(
            name, value, type_, description)


class BoolParameter(SpotlobParameter):
    def __init__(self, name, value):
        super(BoolParameter, self).__init__(name, value, bool, description="")



"""
The pipeline structure is used to define a sequence of processes
to be applied one after another onto a spim, to automate a
detection task consisting of multiple process steps.
"""

import dill


class Pipeline(object):
    """A pipeline is a sequence of processes, that can be applied one after
    another.
    The processes are stored in a Dictionary, along with the SpimStage at which
    they can be applied. The pipeline can be applied completely using the
    `apply_all_steps` method or partially using the `apply_from_stage_to_stage`
    method.
    """

    def __init__(self, processes):
        self.process_stage_dict = dict([(p.input_stage, p) for p in processes])

    @property
    def processes(self):
        return self.process_stage_dict.values()

    def replaced_with(self, new_process):
        """This will give a new pipeline, where one process is replaced with
        the given one

        Parameters
        ----------
        new_process : SpotlobProcessStep
            the new process to be inserted

        Returns
        ----------
        Pipeline
            the pipeline, that includes `new_process`

        """
        new_dict = self.process_stage_dict.copy()
        new_dict[new_process.input_stage] = new_process

        return Pipeline(new_dict.values())

    def apply_from_stage_to_stage(self, spim, from_stage, to_stage):
        """Recursively applies the pipeline-processes from a given stage up to
        another given stage.

        Parameters
        ----------
        spim : Spim
            The image item to apply parts of the pipeline to
        from_stage : int
            SpimStage at which stage the first process should be applied
        to_stage : int
            SpimStage at which stage the last process should be applied

        Returns
        ----------
        Spim
            The processed Spim at stage `to_stage`

        Raises
        ------
        Exception:
            If to_stage is before from_stage

        """
        if from_stage == to_stage:
            return spim
        elif from_stage > to_stage:
            # return self.apply_from_stage_to_stage(spim, to_stage, from_stage)
            raise Exception("invalid apply request")
        else:
            # recursive
            intermediate_spim = self.apply_from_stage_to_stage(
                spim, from_stage, to_stage-1)
            last_process = self.process_stage_dict[to_stage-1]
            outspim = intermediate_spim.do_process_at_stage(last_process)
            return outspim

    def _maxstage(self):
        return max(self.process_stage_dict.keys())+1

    def apply_all_steps(self, spim):
        """Apply the complete pipeline on a given spim

        Parameters
        ----------
        spim : Spim
            the spotlob image item to apply the complete pipeline to

        Returns
        -------
        Spim
            a spotlob image item a the stage after the last process
        """

        minstage = min(self.process_stage_dict.keys())
        maxstage = self._maxstage()
        return self.apply_from_stage_to_stage(spim, minstage, maxstage)

    def apply_at_stage(self, spim):
        """Applies all steps following the stage of the spim

        Parameters
        ----------
        spim : Spim
            the spotlob image item to apply the pipeline to

        Returns
        ----------
        Spim
            the processed Spim at stage `to_stage`
        """

        startstage = spim.stage
        maxstage = self._maxstage()
        return self.apply_from_stage_to_stage(spim, startstage, maxstage)

    def apply_outdated_up_to_stage(self, spim, up_to_stage):
        """Applies all processes since the first outdated one on spim up to
        a given stage if no process is outdated or if the outdated stage
        is past up_to_stage, spim is processed up to up_to_stage,
        or a predecessor is returned at up_to_stage

        Parameters
        ----------
        spim : Spim
            Spim to apply the pipeline to
        up_to_stage : int
            SpimStage up to which the pipeline should be applied

        Returns
        -------
        Spim
            the processed Spim at stage `up_to_stage`
        """
        first_outdated_stage = -1

        # find first stage that is outdated
        for st, process in sorted(self.process_stage_dict.items()):
            if process.outdated:
                first_outdated_stage = st
                break

        if first_outdated_stage >= 0:
            if first_outdated_stage < up_to_stage:
                return self.apply_from_stage_to_stage(
                    spim, first_outdated_stage, up_to_stage)

        # nothing is outdated or outdated stage is past up_to_stage
        if up_to_stage > spim.stage:
            return self.apply_from_stage_to_stage(spim,
                                                  spim.stage,
                                                  up_to_stage)
        else:
            return spim.get_at_stage(up_to_stage)

    def save(self, target_path):
        """Store the pipeline including process paramaters for
        later use.

        Parameters
        ----------
        target_path : str
            path of the file to store the pipeline to

        Notes
        -----
        This creates a file that is also suitable for batch process
        and parallelization.

        See also
        --------
        Pipeline.from_file
            Use `Pipeline.from_file` to restore the pipeline object
            from storage
        """
        with open(target_path, "wb") as dill_file:
            dill.dump(self, dill_file)

    @classmethod
    def from_file(cls, filepath):
        """Restore a pipeline from a file

        Parameters
        ----------
        filepath : str
            filepath of the pipeline file

        Returns
        -------
        Pipepline
            the restored pipeline object
        """

        with open(filepath, "rb") as dill_file:
            restored_pipe = dill.load(dill_file)
        return restored_pipe

    def __str__(self):
        out = []
        for stage in sorted(self.process_stage_dict.keys()):
            process = self.process_stage_dict[stage]
            out += [str(process)]
        return "".join(out)



import numpy as np
import cv2


def distance_point_to_line(x0, y0, pt_on_line1, pt_on_line2):
    """
    Distance from point (x0, y0) to a line defined by two points
    pt_on_line1 and pt_on_line2
    see https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points # noqa
    """

    x1, y1 = pt_on_line1
    x2, y2 = pt_on_line2

    numerator = np.abs((y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1)
    denom = np.sqrt((y2-y1)**2 + (x2-x1)**2)

    return numerator / denom


def max_extends(contours):
    max_extends = np.array([c.max(axis=0) for c in contours])
    return max_extends.max(axis=0)[0]


def points_within_contours(contours):
    """
    gives the indices of points within the given contours
    using the drawContours function
    """

    extend_x, extend_y = max_extends(contours)

    # get maximum points from contours to create mask image
    # of that shape that just covers all

    mask = np.zeros((extend_y, extend_x))

    # fill mask image
    cv2.drawContours(mask, contours, -1, 1, thickness=cv2.FILLED)

    mask = mask.astype(bool)

    yi, xi = np.indices((extend_y, extend_x))
    return np.vstack([xi[mask], yi[mask]]).T


def straight_line_rectangle_collision(line_parameters, rectangle):
    """Calculate the position of the two points where a line crosses the
    boundaries of a rectangle

    Parameters
    ----------
    line_parameters : list of float
        vx,vy, x0,y0
    rectangle : list of float
        corner_x, corner_y, width, height
    """
    x0, y0, vx, vy = line_parameters
    l1, l2 = np.array([x0, y0]), np.array([x0+vx, y0+vy])
    rx0, ry0, w, h = rectangle

    left_border = np.array([rx0, ry0]), np.array([rx0, ry0+h])
    right_border = np.array([rx0+w, ry0]), np.array([rx0+w, ry0+h])
    upper_border = np.array([rx0, ry0+h]), np.array([rx0+w, ry0+h])
    lower_border = np.array([rx0, ry0]), np.array([rx0+w, ry0])

    borders = [left_border, right_border, upper_border, lower_border]

    valid_intersects = []
    for b1, b2 in borders:
        x, y = get_intersection(l1, l2, b1, b2)

        if (rx0 <= x <= rx0+w) and (ry0 <= y <= ry0+h):
            valid_intersects.append(np.array([x, y]))

    assert len(valid_intersects) == 2

    return valid_intersects[0], valid_intersects[1]


def perp(a):
    """perpendicual 2d vector to given vector a"""
    b = np.empty_like(a)
    b[0] = -a[1]
    b[1] = a[0]
    return b


def get_intersection(a1, a2, b1, b2):
    """calculates the intersection point of two line segments

    Parameters
    ----------
    a1 : np.array
        endpoint 1 of segment 1
    a2 : np.array
        endpoint 2 of segment 1
    b1 : np.array
        endpoints 1 of segment 2
    b2 : np.array
        endpoints 2 of segment 2

    Returns
    -------
    np.array
        intersection point
    """
    # https://stackoverflow.com/questions/3252194/numpy-and-line-intersections
    # line segment a given by endpoints a1, a2
    # line segment b given by endpoints b1, b2

    # line segment intersection using vectors
    # see Computer Graphics by F.S. Hill

    da = a2-a1
    db = b2-b1
    dp = a1-b1
    dap = perp(da)
    denom = np.dot(dap, db)
    num = np.dot(dap, dp)
    return (num / denom.astype(float))*db + b1



import cv2
import pandas as pd
import numpy as np
import os

from .process_steps import Reader, Converter, Preprocessor,\
    Binarization, Postprocessor, FeatureFinder, FeatureFilter
from .parameters import SpotlobParameterSet, EnumParameter,\
    BoolParameter, NumericRangeParameter


class SimpleReader(Reader):
    """Reads an image from a file as an RGB file.
    Standard image formats, such as `png`, `jpg`, `tif` are supported.
    It uses `cv2.imread`.
    """

    def __init__(self):
        super(SimpleReader, self).__init__(self.fn_read, [])

    def fn_read(self, filepath):
        # load as color, convert from BGR to RGB
        if os.path.exists(filepath):
            return cv2.cvtColor(cv2.imread(filepath), cv2.COLOR_BGR2RGB), {}
        else:
            raise IOError("File %s not found" % filepath)


class GreyscaleConverter(Converter):
    """Converts a color image to a greyscale image, by selecting one channel
    or by converting it to another color space and then selecting one channel.

    The supported options are given by the `conversion` parameter, which must be
    one of the following strings
    - `Hue`, `Saturation` or `Value` channel
    - `Red`, `Blue` or `Green` color channel
    - normal `Greyscale` conversion

    It uses the `cv2.cvtColor` function.
    Additionally the dark an bright parts can be switched using `invert=True`
    """

    def __init__(self):
        self.hsv_str_list = ["Hue", "Saturation", "Value"]
        self.rgb_str_list = ["Red channel", "Blue channel", "Green channel"]
        converter_options = ["Grey"] + self.hsv_str_list + self.rgb_str_list

        pars = [EnumParameter("conversion",
                              converter_options[0],
                              converter_options),
                BoolParameter("invert", False)]
        super(GreyscaleConverter, self).__init__(self.convert, pars)

    def convert(self, rgb_image, conversion, invert):
        if conversion == "Grey":
            out = cv2.cvtColor(rgb_image, cv2.COLOR_RGB2GRAY).astype(np.uint8)
        else:
            if conversion in self.hsv_str_list:
                code = cv2.COLOR_RGB2HSV
                ch_nr = self.hsv_str_list.index(conversion)
                out = cv2.cvtColor(rgb_image, code)
            elif conversion in self.rgb_str_list:
                ch_nr = self.rgb_str_list.index(conversion)
                out = rgb_image

            out = out[:, :, ch_nr].astype(np.uint8)

        if invert:
            return cv2.bitwise_not(out)
        else:
            return out


class GaussianPreprocess(Preprocessor):
    """Blur the image with a gaussian blur with kernel size given 
    by `ksize`. It uses the `cv2.filter2D` function
    """

    def __init__(self, ksize):
        pars = [NumericRangeParameter("kernelsize", ksize, 1, 47, step=2)]
        super(GaussianPreprocess, self).__init__(self.filter_fn, pars)

    def filter_fn(self, grey_image, kernelsize):
        if kernelsize > 1:
            kernel = np.ones((kernelsize, kernelsize), np.float32)
            kernel /= kernelsize**2
            return cv2.filter2D(grey_image, -1, kernel)
        else:
            return grey_image


class BinaryThreshold(Binarization):
    """Converts the image to a binary one, where the parts above
    the given threshold are set to 255 and the parts below it to 0.
    The sole parameter is the threshold value.
    It uses the `cv2.threshold` function.    
    """

    def __init__(self, threshold):
        pars = SpotlobParameterSet(
            [NumericRangeParameter("threshold", threshold, 0, 255)])
        super(BinaryThreshold, self).__init__(self.threshold_fn, pars)

    def threshold_fn(self, grey_image, threshold):
        _, im = cv2.threshold(
            grey_image, threshold, 255, cv2.THRESH_BINARY)
        return im


class OtsuThreshold(Binarization):
    """Performs a binarization based on Otsu's algorithm.
    It uses the `cv2.threshold` function.    
    """

    def __init__(self):
        super(OtsuThreshold, self).__init__(self.threshold_fn, [])

    def threshold_fn(self, grey_image):
        _, im = cv2.threshold(grey_image,
                              0,
                              255,
                              cv2.THRESH_OTSU)
        return im


class PostprocessNothing(Postprocessor):
    """This process is used as a placeholder for a postprocessing step
    and does not modify the image at all"""

    def __init__(self):
        super(PostprocessNothing, self).__init__(self.postprocess_fn, [])

    def postprocess_fn(self, im):
        return im


class ContourFinderSimple(FeatureFinder):
    """Finds contours, i.e. lists of points that enclose connected areas of
    the same value. It is based on the `cv2.findContours` function
    """

    def __init__(self):
        super(ContourFinderSimple, self).__init__(self.finder_fn, [])

    def finder_fn(self, bin_im):
        cont_ret = cv2.findContours(bin_im,
                                    cv2.RETR_EXTERNAL,
                                    cv2.CHAIN_APPROX_SIMPLE)
        # cont_ret is
        # contours, hierarchy for opencv >4.0
        # im, contours, hierarchy for opencv <=3.4

        contours = cont_ret[-2]
        return contours


class ContourFinder(FeatureFinder):
    """Finds contours, i.e. lists of points that enclose connected areas of
    the same value. It is based on the `cv2.findContours` function.
    It can distinguish between different levels of nested areas

    Parameters
    ----------
    mode : string
        Select which kind of blobs should be found and the contour of
        which should be returned. Select of the following

        * all = all contours, both holes and non-holes

        * inner = innermost blobs without holes in them

        * outer = only outermost blobs

        * holes = only holes, that are contained in other blobs

        * non-holes = all blobs, that are not holes
    """

    def __init__(self, mode):
        mode_par = EnumParameter("mode", mode,
                                 ["outer",
                                  "inner",
                                  "all",
                                  "holes",
                                  "non-holes"])
        super(ContourFinder, self).__init__(self.finder_fn, [mode_par])

    def finder_fn(self, bin_im, mode):
        if mode == "outer":
            cont_ret = cv2.findContours(bin_im,
                                        cv2.RETR_EXTERNAL,
                                        cv2.CHAIN_APPROX_SIMPLE)
            return cont_ret[-2]
        else:
            cont_ret = cv2.findContours(bin_im,
                                        cv2.RETR_CCOMP,
                                        cv2.CHAIN_APPROX_SIMPLE)
            # cont_ret is
            # * contours, hierarchy for opencv >4.0
            # * im, contours, hierarchy for opencv <=3.4

            contours = cont_ret[-2]
            hierarchy = cont_ret[-1][0]

            if mode == 'all':
                return contours
            else:
                holes = []
                nonholes = []
                inner = []

                i = 0

                nonholes_i = []

                # first hierarchy is the nonholes
                # parse through them
                while i >= 0:
                    head = contours[i]
                    nonholes.append(head)
                    nonholes_i.append(i)

                    next_i, prev_i, child_i, parent_i = hierarchy[i]

                    if child_i < 0:
                        inner.append(head)

                    i = next_i  # index of next contour on same hierarchy level

            if mode == "holes":
                # rest should be holes
                # ie. all which are not nonholes
                # list of holes is the contour list withouth the ones
                # specified by nonholes index list
                holes = np.delete(contours, nonholes_i, axis=0)
                return holes
            elif mode == "non-holes":
                return nonholes
            elif mode == "inner":
                return inner
            else:
                raise NotImplementedError(
                    "Unsupported contour finder mode %s" % mode)


class FeatureFormFilter(FeatureFilter):
    """It analyzes the contours and filters them using given criteria:

    - the enclosed area must be smaller (i.e. contain fewer pixels) than
      `minimal_area`
    - it solidity, i.e. the ratio of the area of the contour and its convex
      hull must be below a given value
    - if `remove_on_edge` is `True`, contours that touch the border of the
      image are filtered out

    """

    def __init__(self, size, solidity, remove_on_edge):
        pars = [NumericRangeParameter("minimal_area", size, 0, 10000),
                NumericRangeParameter("solidity_limit",
                                      solidity, 0, 1, step=0.01, type_=float),
                BoolParameter("remove_on_edge", remove_on_edge)]
        super(FeatureFormFilter, self).__init__(self.filter_fn, pars)

    def solidity(self, c):
        try:
            return cv2.contourArea(c)/cv2.contourArea(cv2.convexHull(c))
        except ZeroDivisionError:
            return 0

    def filter_fn(self,
                  contours,
                  image_shape,
                  minimal_area,
                  solidity_limit,
                  remove_on_edge):

        def valid_contour(c):
            valid_area = cv2.contourArea(c) > minimal_area
            valid_solidity = self.solidity(c) > solidity_limit
            valid_on_edge = not remove_on_edge or \
                self.is_off_border(c, image_shape)
            return (valid_area and
                    valid_solidity and
                    valid_on_edge)

        return [c for c in contours if valid_contour(c)]

    def is_off_border(self, contour, image_shape):
        """
        this function checks if a contour is touching the border
        of an image shaped like image_shape
        """
        bb_x, bb_y, bb_w, bb_h = cv2.boundingRect(contour)
        maxh = image_shape[0]
        maxw = image_shape[1]

        xMin = 0
        yMin = 0
        xMax = maxw - 1
        yMax = maxh - 1

        if any([bb_x <= xMin,
                bb_y <= yMin,
                bb_x+bb_w >= xMax,
                bb_y+bb_h >= yMax]):
            return False
        else:
            return True

    def draw_contours(self, image, contours):
        background = cv2.cvtColor(image.copy(), cv2.COLOR_GRAY2RGB)
        return draw_contours(background, contours)


def draw_contours(color_image, contours, color=(0, 255, 0), thickness=1):
    return cv2.drawContours(color_image, contours, -1, color, thickness)


def crop_to_contour(image, contours):
    x, y, w, h = cv2.boundingRect(contours)

    return image[y:y+h, x:x+w]



"""
Store results and images from a detection into files
"""
import os.path

from PIL import Image

from .spim import Spim, SpimStage
from .process_opencv import draw_contours


class Writer(object):
    def __init__(self, image_filepath, data_filepath):
        self.image_filepath = image_filepath
        self.data_filepath = data_filepath

    def store_image(self, image, contours):
        image_w_contours = draw_contours(image, contours)
        im = Image.fromarray(image_w_contours)
        im.save(self.image_filepath)
        return self.image_filepath

    def store_data(self, dataframe):
        _, ext = os.path.splitext(self.data_filepath)

        if ext == ".csv":
            dataframe.to_csv(self.data_filepath)
        return self.data_filepath



import json

import cv2
import pandas as pd
import numpy as np

from .process_steps import Analysis
from .parameters import SpotlobParameterSet
from .calculation import points_within_contours, max_extends,\
    distance_point_to_line, straight_line_rectangle_collision, perp
from .process_opencv import draw_contours


class LineAnalysis(Analysis):
    def __init__(self, calibration=None,
                 linewidth_percentile=95,
                 extended_output=True):
        self.calibration = calibration
        self.linewidth_percentile = linewidth_percentile
        super(LineAnalysis, self).__init__(
            self.analyze, [], extended_output=extended_output)

    def analyze(self, metadata):
        contours = metadata['contours']

        if len(contours) == 0:
            empty_df = pd.DataFrame([], columns=["area_px2",
                                                 "linewidth_px",
                                                 "linewidth2_px",
                                                 "bb_width_px",
                                                 "bb_height_px",
                                                 "bb_angle",
                                                 "distances_hist",
                                                 "distances_bin_edges_px"])
            return empty_df

        elif len(contours) == 1:
            inner_points = points_within_contours(contours)
        else:
            inner_points = np.vstack([points_within_contours([ctr])
                                      for ctr in contours])

        area = len(inner_points)

        vx, vy, x0, y0 = cv2.fitLine(inner_points,
                                     cv2.DIST_FAIR, 0, 0.01, 0.01)
        # x0, y0, vx, vy
        line_parameters = x0[0], y0[0], vx[0], vy[0]

        # calculate crossings with borders
        width, height = max_extends(contours)

        image_rectangle = 0, 0, width, height
        p1, p2 = straight_line_rectangle_collision(line_parameters,
                                                   image_rectangle)

        distances = distance_point_to_line(inner_points[:, 0],
                                           inner_points[:, 1],
                                           p1,
                                           p2)

        linewidth_perc = np.percentile(distances, self.linewidth_percentile)*2

        line_length = np.linalg.norm(np.array(p2)-np.array(p1))
        linewidth_shading = area/line_length

        res_dict = {"area_px2": area,
                    "linewidth_px": linewidth_perc,
                    "linewidth_shading_px": linewidth_shading,
                    "line_params": [np.array(line_parameters)],
                    "line_start": [np.array(p1)],
                    "line_end": [np.array(p2)]}

        if self.extended_output:
            hist, bin_edges = np.histogram(distances, bins="auto")
            res_dict.update({"distances_hist": [hist],
                             "distances_bin_edges_px": [bin_edges],
                             "contours": [contours]})

        result = pd.DataFrame(res_dict,
                              index=[0])

        if not self.calibration:
            return result
        else:
            return self.calibration.calibrate(result)

    def draw_results(self, image, dataframe, crop_to_contours=False):
        if len(dataframe) == 1:
            row = dataframe.iloc[0]

            x0, y0, vx, vy = row["line_params"]
            linewidth = row["linewidth_px"]
            linewidth_shading = row["linewidth_shading_px"]

            if "contours" in dataframe:
                contour = row["contours"]
                draw_contours(image, contour)

            cstart = np.round(row["line_start"]).astype(int)
            cstop = np.round(row["line_end"]).astype(int)

            # center line
            cv2.line(image, tuple(cstart), tuple(cstop),
                     (255, 0, 0), 1, lineType=cv2.LINE_AA)
            cv2.circle(image, tuple(cstart), 4, (255, 0, 0), -1)
            cv2.circle(image, tuple(cstop), 4, (255, 0, 0), -1)

            # border lines

            # percentile linewidth
            self._draw_line_borders(image,
                                    linewidth,
                                    (255, 0, 0),
                                    cstart,
                                    cstop,
                                    (vx, vy))

            # # shading linewidth
            self._draw_line_borders(image,
                                    linewidth_shading,
                                    (0, 200, 200),
                                    cstart,
                                    cstop,
                                    (vx, vy))

        return image

    def _draw_line_borders(self,
                           image,
                           linewidth,
                           color,
                           center_start,
                           center_stop,
                           vector):

        # orthogonal vector to line vector with length linewidth
        # assuming vector is normalized to length 1
        orthogonal_v = np.round(perp(vector)*linewidth/2.0).astype(int)

        lower_start = tuple(np.subtract(center_start, orthogonal_v))
        lower_stop = tuple(np.subtract(center_stop, orthogonal_v))
        upper_start = tuple(np.add(center_start, orthogonal_v))
        upper_stop = tuple(np.add(center_stop, orthogonal_v))

        cv2.line(image, lower_start, lower_stop,
                 color, 2, lineType=cv2.LINE_AA)
        cv2.line(image, upper_start, upper_stop,
                 color, 2, lineType=cv2.LINE_AA)



"""A Spim is the object holding the images and metadata.
It has methods, that return a Spim of the next stage.
For example, a blank, empty Spim can be created and is then
in the stage `SpimStage.new`. It contains only the information
where to find the image file. If `Spim.read(Writer)` is called,
a new Spim is returned, which contains the image data and is at
stage `SpimStage.loaded`.

Here is a list of the stages that a Spim can be in and in between,
the methods that return a Spim of the next stage.

.. graphviz::

    strict digraph {
        node [shape=box, width=2]

        0 [label="new", target="_top"];
        1 [label="loaded"];
        2 [label="converted"];
        3 [label="preprocessed", below=2];
        4 [label="binarized", below=1];
        5 [label="postprocessed", below=0];
        6 [label="features_extracted"];
        7 [label="features_filtered"];
        8 [label="analyzed"];
        9 [label="stored"];

        {rank=same;
            0 -> 1 [label="read"];
            1 -> 2 [label="convert"];
        }
        2 -> 3 [label="preprocess"];
        {rank=same;
            4 -> 3 [label="binarize", dir="back"];
            5 -> 4 [label="postprocess", dir="back"];
        }
        5 -> 6 [label="extract_features"];
        {rank=same;
            6 -> 7 [label="filter_features"];
            7 -> 8 [label="analyze"];
        }
        8 -> 9 [label="store"];
    }

With every step, information is collected. A spim at a later stage
does not duplicate the image data from former stages. However, if this
data is still needed, it can contain a reference to its predecessors.
"""


import pandas


class SpimStage(object):
    """Enumeration of the stages that a Spim can go through"""

    new = 0
    loaded = 1
    converted = 2
    preprocessed = 3
    binarized = 4
    postprocessed = 5
    features_extracted = 6
    features_filtered = 7
    analyzed = 8
    stored = 9


class Spim(object):
    """Spotlob image item"""
    # TODO: describe nature of Spim, immutable concept

    def __init__(self, image, metadata, stage, cached, predecessors):
        """A Spim is a **Spotlob image item**, an object representing an image
        and the metadata that is collected along the process through a
        pipeline.

        Parameters
        ----------
        image : numpy array
            an image
        metadata : dict
            the data desribing the image and containing results
        stage : SpimStage
            the stage along the pipeline the image has passed
        cached : bool
            if this is true, a reference to predecessors of this Spim are
            stored and they are kept in memory. This is required if a process
            step is to be repeated
        predecessors : dict(SpimStage, Spim)
            a registry of predecessors of the current spim, stored alongside
            the stage they are in
        """

        self._image = image
        self.metadata = metadata
        self.stage = stage
        self.cached = cached
        self.predecessors = predecessors

        if image is not None:
            self.metadata.update({"image_shape": image.shape})

    @classmethod
    def from_file(cls, image_filepath, cached=False):
        """Create a Spim object from an image file. The path is stored in the
        Spim object, but the image is not yet loaded.

        Parameters
        ----------
        image_filepath : str
            Path to an image file. The image type must be understood by the
            reader that is given when the `read`-function is called. If an
            invalid image type is given at this stage, it will not be
            recognized
        cached : bool, optional
            If the spim is to be cached, a reference to predecessors will be
            kept and not be deleted by the garbage collector. This allows to
            go back to an earlier stage after applying processes, but is more
            memory consuming. (the default is False)

        Returns
        -------
        Spim
            An empty Spim at SpimStage.new, that does not contain any data
            except the filepath
        """

        md = {"filepath": image_filepath}
        return Spim(None,
                    md,
                    SpimStage.new,
                    cached=cached,
                    predecessors=dict())

    @property
    def image(self):
        """Gives the image contained in this Spim or in the latest
        predecessor, that has an image

        Raises
        ------
        Exception
            Exception is raised if no image is present, most likely
            because it has not been cached

        Returns
        -------
        numpy.array
            latest image
        """

        if not (self._image is None):
            return self._image
        elif self.cached:
            return self.predecessor_image()
        else:
            raise Exception("image not found, has not been cached")

    def predecessor_image(self):
        predecessor_stages = self.predecessors.keys()
        predecessor_stages = sorted(predecessor_stages)

        for i in predecessor_stages[::-1]:
            p = self.predecessors[i]
            if not (p.image is None):
                return p.image
        raise Exception("no image found")

    def read(self, reader):
        im, metadata = reader.apply(self.metadata["filepath"])
        metadata.update(self.metadata)
        metadata.update({"image_shape": im.shape})
        return Spim(im,
                    metadata,
                    SpimStage.loaded,
                    self.cached,
                    self._predecessors_and_self())

    def apply_process(self, process):
        assert self.stage == process.input_stage
        im = process.apply(self.image)
        return Spim(im,
                    self.metadata.copy(),
                    self.stage + 1,
                    self.cached,
                    self._predecessors_and_self())

    def convert(self, converter):
        return self.apply_process(converter)

    def preprocess(self, preprocessor):
        return self.apply_process(preprocessor)

    def binarize(self, binarizer):
        return self.apply_process(binarizer)

    def postprocess(self, postprocessor):
        return self.apply_process(postprocessor)

    def extract_features(self, feature_extractor):
        contours = feature_extractor.apply(self.image)
        new_metadata = self.metadata.copy()
        new_metadata.update({"contours": contours})
        newspim = Spim(None, new_metadata, SpimStage.features_extracted,
                       self.cached, self._predecessors_and_self())
        return newspim

    def filter_features(self, feature_filter):
        filtered_contours = feature_filter.apply(self.metadata["contours"],
                                                 self.metadata["image_shape"])
        metadata = self.metadata.copy()
        metadata["contours"] = filtered_contours
        return Spim(None,
                    metadata,
                    SpimStage.features_filtered,
                    self.cached,
                    self._predecessors_and_self())

    def analyze(self, analysis):
        results = analysis.apply(self.metadata)
        metadata = self.metadata.copy()
        metadata["results"] = results
        return Spim(None,
                    metadata,
                    SpimStage.analyzed,
                    self.cached,
                    self._predecessors_and_self())

    def store(self, writer):
        assert self.stage == SpimStage.analyzed

        metadata = self.metadata.copy()

        contours = metadata["contours"]

        fresh_image = self.get_at_stage(SpimStage.loaded).image
        image_path = writer.store_image(fresh_image, contours)
        data_path = writer.store_data(self.get_data())

        metadata["output_image_filepath"] = image_path
        metadata["output_data_path"] = data_path

        return Spim(None,
                    metadata,
                    SpimStage.stored,
                    self.cached,
                    self._predecessors_and_self())

    def func_at_stage(self, spimstage):
        """The method like `self.read()`, `self.convert()`,... that can
        be safely called at the given stage

        Parameters
        ----------
        spimstage : int
            SpimStage that the requested method corresponds to

        Returns
        -------
        callable
            the function, that can be applied the given stage
        """

        # TODO: the static map of functions should be defined elsewhere
        functions = [self.read,
                     self.convert,
                     self.preprocess,
                     self.binarize,
                     self.postprocess,
                     self.extract_features,
                     self.filter_features,
                     self.analyze,
                     self.store]
        return functions[spimstage]

    def do_process_at_stage(self, process):
        """Apply the given process at at this Spim if the process fits
        this stage or at a predecessor of this Spim that fits the
        process' input stage

        Parameters
        ----------
        process : SpotlobProcessStep
            Process to apply

        Returns
        -------
        Spim
            The Spim that results from the process being applied. It is
            in stage `process.input_stage + 1`
        """
        return self.func_at_stage(process.input_stage)(process)

    def _predecessors_and_self(self):
        if self.cached:
            outd = dict()
            for p_stage, p_spim in self.predecessors.items():
                if p_stage < self.stage:
                    outd.update({p_stage: p_spim})
            outd.update({self.stage: self})
            return outd
        else:
            # TODO: should this return self
            return dict()

    def get_at_stage(self, spimstage):
        """Get the Spim at a given stage. This returns a predecessor if it has
        been chached

        Parameters
        ----------
        spimstage : int
            That the returned Spim should be at

        Raises
        ------
        Exception
            If there is no predecessor at the requested stage, for example if
            Spim has not been cached

        Returns
        -------
        Spim
            The Spim at the requested Stage
        """

        if spimstage == self.stage:
            return self
        else:
            try:
                return self.predecessors[spimstage]
            except KeyError:
                # TODO: check if cached = False, then predecessor cannot exist
                msg = "Spim has no predecessor at stage %s." % spimstage
                # TODO: more specific exception predecessor does not exist
                raise Exception(msg)

    def get_data(self):
        """get all metadata and results as flat metadata

        RETURNS
        -------
            pandas.Dataframe
                all metadata including collected results
        """
        # TODO: better tests for get_data

        if "results" in self.metadata.keys():
            # results is a dataframe
            # flatten to one dataframe

            md_copy = self.metadata.copy()

            results = md_copy.pop("results")
            results["filepath"] = md_copy["filepath"]

            # TODO: find a way to include contours output in get_data
            # drop contours
            _ = md_copy.pop("contours")

            md = pandas.Series(md_copy).to_frame().T

            return results.merge(md, on="filepath")
        else:
            return pandas.Series(self.metadata).to_frame().T

    def __repr__(self):
        return "<Spim instance %s at stage %s>" % (id(self), self.stage)



import matplotlib.pyplot as plt


class MatplotlibPreviewScreen(object):
    def __init__(self):
        self.fg_ax = None

    def make_new(self, background_im, *args, **kwargs):
        kwargs.update({'num': 'spotlob'})

        if "figsize" not in kwargs.keys():
            h, w = background_im.shape[:2]
            aspect = h/w
            width = 7
            kwargs.update({"figsize": (width, aspect*width)})

        self.fig = plt.figure(*args, **kwargs)
        self.ax = plt.Axes(self.fig, [0., 0., 1., 1.])
        self.fig.add_axes(self.ax)
        # self.bg_ax = self.ax.imshow(background_im)
        self.fg_ax = self.ax.imshow(background_im, cmap="gray")
        self.ax.set_axis_off()

    def update(self, new_im):
        """redraw the image if foreground axis is defined"""
        if self.fg_ax:
            self.fg_ax.set_data(new_im)
            self.fig.canvas.draw_idle()



"""
The detection of features within an image with spotlob is split up into
an abstract but fixed sequence of processes. Any of these process steps is
applied onto a Spim and returns a new Spim. The new Spim contains the
information added by the process step.

.. graphviz::

    digraph seq {
        rankdir="LR";
        "spim1" -> "spim2" [label="spim1.function(process)"];
    }

For example, a `Spim` at stage `SpimStage.loaded` can be converted, using
a concrete subclass of `Converter`, named `process_opencv.GreyscaleConverter`
in the following way:

.. code-block:: python

    from spotlob.defaults import load_image
    from spotlob.process_opencv import GreyscaleConverter

    loaded_spim = load_image("color_image.jpg")
    my_converter = GreyscaleConverter()
    converted_spim = loaded_spim.convert(my_converter)
"""


import numpy as np

from .spim import SpimStage
from .parameters import SpotlobParameterSet


class SpotlobProcessStep(object):
    """An abstract super class for process steps. A process step can be
    applied to bring a spim from one stage to another. It is supposed to save
    as internal state, if it has already been applied or needs to be applied
    again, because the parameters have changed. This is stored in the
    `outdated` parameter"""

    # stage that a spim should have before this Process can be applied
    input_stage = None

    # a process is outdated, if it has not been
    # applied since the parameters changed
    outdated = True

    def __init__(self, function, parameters, add_to_register=True):
        """
        A process step can be applied to bring a spim from one stage
        to another

        Parameters
        ----------
        function : callable
            a function to be applied on spim with optional additional
            parameters
        parameters : list of SpotlobParameter
            the parameters that will be used for the function as soon as
            it is applied
        add_to_register : bool, optional
            if this is True, the process will be registered centrally now, i.e.
            upon creation of the SpotlobProcessStep object. This allows the
            user to know which processes are available (the default is True)

        """
        self.function = function
        self.parameters = SpotlobParameterSet(parameters)

        # import needs to be delayed to avoid circular imports
        from .register import PROCESS_REGISTER

        if add_to_register:
            PROCESS_REGISTER.register_process(self)

    def preview(self, spim):
        """
        This function takes spim at an undefined stage and draws
        the effect of the process on top, to provide a preview for the
        user on how the funciton will work. No storage or sideeffects
        should take place.
        In contrast to the apply function it must always return an image

        Parameters
        ----------
        spim : Spim
            Spim to draw the preview on. It must contain an image

        """
        raise NotImplementedError("abstract: to be implemented by subclass")

    def apply(self, *input_args):
        """Apply the process step onto input data (eg. images or contours)
        and yield the output data (again, could be a modified image or gathered
        data). The actual return values depend on the concrete implementation
        of the function. Additional arguments for the function are the
        parameters of the process as stored in the `parameters` field of each
        process.

        Returns
        -------
        Contours or images
            Output of the concrete process implementation
        """

        output = self.function(*input_args, **self.parameters.to_dict())
        self.outdated = False
        return output

    def __str__(self):
        classname = str(type(self).__name__)
        parameters_str = str(self.parameters)
        return "%s\n%s" % (classname, parameters_str)



import numpy as np

from ipywidgets import IntSlider, FloatSlider, Text,\
    Dropdown, VBox, Checkbox, Button, Output
from IPython.display import display, clear_output

from .parameters import EnumParameter, BoolParameter
from .spim import Spim, SpimStage


class SpotlobNotebookGui(object):
    def __init__(self, pipeline, preview_screen, spim):
        self.pipeline = pipeline
        self.dummyspim = spim  # Spim.from_file(image_filepath, cached=True)
        self.preview_screen = preview_screen

    # def new_image(self, image_filepath):
    #     self.dummyspim = Spim.from_file(image_filepath, cached=True)
    #     self.run()

    def run(self):
        self.dummyspim = self.pipeline.apply_all_steps(self.dummyspim)
        self.update_preview(self.pipeline.process_stage_dict[SpimStage.loaded])

    def update_preview(self, process_that_changed):
        spim_before = self.dummyspim.get_at_stage(
            process_that_changed.input_stage)

        preview_image = process_that_changed.preview(spim_before)
        self.preview_screen.update(preview_image)

    def make_widgets(self):
        self.run()
        return self.pipeline_as_widget(self.pipeline)

    def show_preview_screen(self, *args, **kwargs):
        self.preview_screen.make_new(
            self.dummyspim.get_at_stage(SpimStage.loaded).image,
            *args, **kwargs)

    def parameter_as_widget(self, parameter, parent_process):
        v = parameter.value
        ty = parameter.type
        name = parameter.name

        # if isinstance(parameter, FilepathParameter):
        #     return fileselect.SelectFilesButton()
        # TODO: handle other than a string
        try:
            min_ = parameter.minvalue
            max_ = parameter.maxvalue
            step = parameter.step

            # use a slider
            if ty == float:
                widget_class = FloatSlider
            elif ty == int:
                widget_class = IntSlider
            else:
                NotImplementedError("Unsupported parameter type")

            widget = widget_class(min=min_, max=max_,
                                  step=step, value=v,
                                  description=name,
                                  continuous_update=False)

        except AttributeError:
            if isinstance(parameter, EnumParameter):
                widget = Dropdown(options=parameter.options,
                                  value=parameter.value, description=name)
            elif isinstance(parameter, BoolParameter):
                widget = Checkbox(value=parameter.value,
                                  description=name)
            elif ty == str:
                widget = Text(value=v, description=name, disabled=False)
            else:
                raise NotImplementedError(
                    "Could not construct widget for parameter type %s" % ty)

        def update_to_widgetstate(widgetstate):
            parameter.value = widgetstate["new"]

            # redo the whole invalidated pipeline up to the parent process
            self.dummyspim = self.pipeline.apply_outdated_up_to_stage(
                self.dummyspim, parent_process.input_stage)
            # update the preview of the process
            self.update_preview(parent_process)

            # mark process as outdated --> it has to be reapplied for others
            parent_process.outdated = True

        # bind functionality to widget ###########
        widget.observe(update_to_widgetstate, names="value")

        return widget

    def pipeline_as_widget(self, pipeline):
        return VBox([
            VBox([self.parameter_as_widget(p, process)
                  for p in process.parameters.parameters])
            for process in pipeline.processes if process
        ])

    def run_button(self):
        rb = Button(description="Evaluate")

        def run_inner(btn):
            self.run()

            fresh_predecessor = self.dummyspim.get_at_stage(SpimStage.loaded)

            # draw results
            image = np.copy(fresh_predecessor.image)

            before_analyzed = SpimStage.analyzed-1

            analyis_process = self.pipeline.process_stage_dict[before_analyzed]
            image_with_results = analyis_process.draw_results(
                image, self.results())
            self.preview_screen.update(image_with_results)

        rb.on_click(run_inner)
        return rb

    def results(self):
        return self.dummyspim.metadata["results"]

    def results_button(self):
        rb = Button(description="Show Results")
        out = Output()

        def results_button_inner(btn):
            out.clear_output()
            # display(rb)
            with out:
                display(self.results())

        rb.on_click(results_button_inner)
        # return rb
        display(rb)
        display(out)



import json

import cv2
import pandas as pd
import numpy as np

from .process_steps import Analysis
from .parameters import SpotlobParameterSet
from .process_opencv import draw_contours, crop_to_contour


class CircleAnalysis(Analysis):
    def __init__(self, calibration=None, extended_output=True):
        self.calibration = calibration
        super(CircleAnalysis, self).__init__(
            self.analyze, [], extended_output=extended_output)

    def analyze(self, metadata):
        contours = metadata['contours']
        areas = []
        ellipses_positions = []
        ellipses_major_axes = []
        ellipses_minor_axes = []
        ellipses_angles = []

        for cont in contours:
            # AREA
            areas += [cv2.contourArea(cont)]

            # ELLIPSE
            try:
                e_pos, (e_major_ax, e_minor_ax), angle = cv2.fitEllipse(cont)
            except cv2.error:
                e_pos, (e_major_ax,
                        e_minor_ax), angle = np.nan, (np.nan, np.nan), np.nan

            ellipses_positions += [np.array(e_pos)]
            ellipses_major_axes += [e_major_ax]
            ellipses_minor_axes += [e_minor_ax]
            ellipses_angles += [angle]

        res_dict = {"area_px2": areas,
                    "ellipse_position_px": ellipses_positions,
                    "ellipse_majorAxis_px": ellipses_major_axes,
                    "ellipse_minorAxis_px": ellipses_minor_axes,
                    "ellipse_angle": ellipses_angles}

        if self.extended_output:
            res_dict.update({"contours": contours})

        result = pd.DataFrame(res_dict)

        if not self.calibration:
            return result
        else:
            return self.calibration.calibrate(result)

    def draw_results(self, image, dataframe, crop_to_contours=False):
        for _, row in dataframe.iterrows():
            pos_x, pos_y = row["ellipse_position_px"]
            e_major = row["ellipse_majorAxis_px"]
            e_minor = row["ellipse_minorAxis_px"]
            angle = row["ellipse_angle"]

            e_pos = (int(pos_x), int(pos_y))
            e_size = (int(e_major/2.0), int(e_minor/2.0))

            pen_color = [255, 0, 0]

            if "contours" in row.keys():
                contour = row["contours"]
                draw_contours(image, contour)

            cv2.circle(image, e_pos, 3, pen_color, -1)
            cv2.ellipse(image, e_pos, e_size, angle, 0, 360, pen_color, 1)

            if crop_to_contours:
                if "contours" in row.keys():
                    contour = row["contours"]
                    image = crop_to_contour(image, contour)

        return image



import multiprocessing as mp
import warnings

import pandas as pd

from .spim import Spim
from .pipeline import Pipeline


def is_interactive():
    """checks wether called in an interactive environment"""
    import __main__ as main
    return not hasattr(main, '__file__')


def _process_job(job):
    # function cannot be pickled,
    # so load function with dill within function call
    pipeline_file, image_file = job
    pipeline = Pipeline.from_file(pipeline_file)
    myspim = Spim.from_file(image_file)
    processed_spim = pipeline.apply_at_stage(myspim)
    return processed_spim.get_data()


def batchprocess(pipeline_file, image_files, multiprocessing=False):
    """This function applies a pipeline from a file onto a stack of images.
    The results are collected in one :class:`pandas.Dataframe`.

    PARAMETERS
    ----------
    pipeline_file : str
        the filepath of a pickled pipeline
    image_files : list of str
        paths of the images
    multiprocessing : bool, optional
        if True, the processing will be done in parallel using multiple cpu
        cores at once.
    
    RETURNS
    -------
    pandas.Dataframe
        Flat Dataframe where one row corresponds to one detected feature
    """
    if multiprocessing:
        if is_interactive():
            warnings.warn(
                """It seems you are running in an interactive environment.
                Multiprocessing might not work.
                Consider using ipyparallel instead""")

        jobs = zip([pipeline_file]*len(image_files), image_files)
        no_cores = mp.cpu_count()
        pool = mp.Pool(processes=no_cores)
        try:
            res = pool.map(_process_job, jobs)
        finally:
            pool.close()
            pool.join()

    else:
        pipeline = Pipeline.from_file(pipeline_file)

        def process_file(fn):
            myspim = Spim.from_file(fn)
            respim = pipeline.apply_at_stage(myspim)
            return respim.get_data()

        res = map(process_file, image_files)
    #
    # make large dataframe out of list of dataframes
    return pd.concat(res)



"""The register is meant to be used to keep track of available
process steps. Using decorators, a function can be internally
turned into a :class:`~spotlob.SpotlobProcessStep` subclass, to be used
within a :class:`~spotlob.Pipeline`. This way, using only minimal code,
new functionality can be added to Spotlob and directly used
within its workflow

Example using a decorator
-------------------------
Create a process register

.. code-block:: python

    from spotlob.register import ProcessRegister
    register = ProcessRegister()

Create an alternative function, that should replace a single
:class:`~spotlob.SpotlobProcessStep`. Here, a binarization function is defined,
using upper and lower value boundaries, with `numpy`.

.. code-block:: python

    import numpy as np

    def my_threshold(image, lower_threshold, upper_threshold, invert):
        out = np.logical_and( image > lower_threshold,
                            image < upper_threshold)
        out = out.astype(np.uint8)*255
        if invert:
            out = ~out
        return out

To add this function to the register, use the methods of
:class:`~spotlob.ProcessRegister` as decorators, giving a list of parameter
specifications

.. code-block:: python

    @register.binarization_plugin([("lower_threshold",(0,255,100)),
                                   ("upper_threshold",(0,255,200)),
                                   ("invert", True)])
    def my_threshold(image, lower_threshold, upper_threshold, invert):
        out = np.logical_and( image > lower_threshold,
                            image < upper_threshold)
        out = out.astype(np.uint8)*255
        if invert:
            out = ~out
        return out
"""

from .parameters import parameter_from_spec
from .process_steps import Binarization, Converter, Preprocessor,\
    Postprocessor, Analysis


class ProcessRegister(object):

    def __init__(self):
        self.available_processes = dict()

    def binarization_plugin(self, param_spec):
        """
        register your binarization function as a plugin by using
        this function as a decorator
        """
        wrapper = self._get_wrapper(Binarization, param_spec)
        return wrapper

    def convert_plugin(self, param_spec):
        wrapper = self._get_wrapper(Converter, param_spec)
        return wrapper

    def preprocess_plugin(self, param_spec):
        wrapper = self._get_wrapper(Preprocessor, param_spec)
        return wrapper

    def postprocess_plugin(self, param_spec):
        wrapper = self._get_wrapper(Postprocessor, param_spec)
        return wrapper

    def analysis_plugin(self, param_spec):
        wrapper = self._get_wrapper(Analysis, param_spec)
        return wrapper

    def _get_wrapper(self, process_class_name, process_param_spec):
        def wrapper(process_function):
            # create SpotlobParameters out of definition in decorator
            spotlob_parameters = [parameter_from_spec(
                spec) for spec in process_param_spec]

            # instantiate given class name, handing over parameters,
            # do not register the object
            proc = process_class_name(
                process_function, spotlob_parameters, add_to_register=False)

            # register the object now, but via the name of the function
            self.available_processes[process_function.__name__] = proc
            return process_function

        return wrapper

    def register_process(self, process_instance):
        self.available_processes[process_instance.__class__] = process_instance


PROCESS_REGISTER = ProcessRegister()



__name__ = "spotlob"
__version__ = "0.9.1.a"
__author__ = "Fabian Meyer"
__summary__ = "feature extraction and analysis pipeline for image data"
__copyright__ = "Copyright 2020, Fraunhofer ISE"
__license__ = "BSD 3-clause"
__maintainer__ = "Fabian Meyer"
__email__ = "fabian.meyer@ise.fraunhofer.de"
__url__ = "https://gitlab.cc-asp.fraunhofer.de/fmeyer/spotlob"



"""For every function of `Spim` that returns another `Spim` at
a further stage, there is a subclass of `SpotlobProcessStep`,
that can be used as super class for an concrete implementation
of that step. The return type of a `SpotlobProcessStep.apply`
call is different depending on the type of process. The Spim
internally passes the modified data to the new Spim created
through the process.
"""

from .spim import Spim, SpimStage
from .process import SpotlobProcessStep


class Reader(SpotlobProcessStep):
    """A reader loads the image data from storage into memory.
    `apply` returns an image"""
    input_stage = SpimStage.new


class Converter(SpotlobProcessStep):
    """A converter converts a color image to a greyscale image.
    `apply` returns a greyscale image
    """
    input_stage = SpimStage.loaded

    def preview(self, spim):
        input = spim.get_at_stage(self.input_stage).image
        return self.apply(input)


class Preprocessor(SpotlobProcessStep):
    """Preprocessing is applied onto a grey image to prepare for
    binarization, for example by cleaning the image from unwanted
    features.
    `apply` returns a greyscale image
    """
    input_stage = SpimStage.converted

    def preview(self, spim):
        input = spim.get_at_stage(self.input_stage).image
        return self.apply(input)


class Binarization(SpotlobProcessStep):
    """Turns a greyscale image into a black-and-white or binary image.
    `apply` returns a greyscale image
    """
    input_stage = SpimStage.preprocessed

    def preview(self, spim):
        input = spim.get_at_stage(self.input_stage).image
        binim = self.apply(input)
        # bin_masked = ~np.ma.masked_array(binim, binim == 0)
        # return bin_masked
        return binim


class Postprocessor(SpotlobProcessStep):
    """Postprocessing is done on a binary image to facilitate the
    detection.
    `apply` returns a greyscale image
    """
    input_stage = SpimStage.binarized

    def preview(self, spim):
        input = spim.get_at_stage(self.input_stage).image
        return self.apply(input)


class FeatureFinder(SpotlobProcessStep):
    """The FeatureFinder tries to find contours in a binary image.

    `apply` returns contours
    """
    input_stage = SpimStage.postprocessed

    def preview(self, spim):
        # do nothing, get at stage before feature find
        return spim.get_at_stage(self.input_stage).image


class FeatureFilter(SpotlobProcessStep):
    """The FeatureFilter can reduce the number of detected features
    by analyzing them.

    `apply` returns contours
    """
    input_stage = SpimStage.features_extracted

    def preview(self, spim):
        input_ = spim.get_at_stage(self.input_stage)
        new_contours = self.apply(input_.metadata["contours"],
                                  input_.metadata["image_shape"])

        background = spim.get_at_stage(SpimStage.binarized).image
        return self.draw_contours(background, new_contours)

    def draw_contours(self, image, contours):
        raise NotImplementedError("abstract: to be implemented by subclass")


class Analysis(SpotlobProcessStep):
    """An Analysis class evaluates the metadata (including contours) and 
    yields its results as a dataframe.

    `apply` returns :class:`~pandas.DataFrame`"""
    input_stage = SpimStage.features_filtered

    def __init__(self,
                 function,
                 parameters,
                 add_to_register=True,
                 extended_output=True):
        """Abstract analysis process step. 

        Parameters
        ----------
        function : callable
            function that performs the analysis. As first argument it must
            take the metadata dict and further arguments must match the
            Parameters of the SpotlobParameterSet.
            Must return a pandas.Dataframe
        parameters : SpotlobParameterSet
            Parameters of this process steps
        add_to_register : bool, optional
            wether this should automatically be added to the list of known
            processes, by default True
        extended_output : bool, optional
            if the output of this Analysis should contain as much data as
            possible. If False, only most important results are returned. 
            By default True
        """
        self.extended_output = extended_output
        super(Analysis, self).__init__(function, parameters, add_to_register)

    def preview(self, spim):
        df = self.apply(spim.metadata)

        im = spim.get_at_stage(SpimStage.loaded).image
        return self.draw_results(im, df)

    def draw_results(self, image, dataframe, crop_to_contours=False):
        raise NotImplementedError("abstract: to be implemented by subclass")



from IPython.display import display

import os.path

from .widget import SpotlobNotebookGui
from .process_opencv import SimpleReader, GreyscaleConverter,\
    GaussianPreprocess, OtsuThreshold, BinaryThreshold, PostprocessNothing, \
    ContourFinderSimple, FeatureFormFilter
from .pipeline import Pipeline
from .preview import MatplotlibPreviewScreen
from .register import PROCESS_REGISTER
from .analyze_circle import CircleAnalysis
from .analyze_line import LineAnalysis
from .spim import Spim


def default_pipeline(mode="circle", thresholding="auto"):
    """Gives a pipeline which works for many cases and can be used as a
    starting point for further tuning or as default for the GUI notebook.
    By default, features with an area smaller than 500 pixels are ignored.

    Parameters
    ----------
    mode : str, optional
        this defines the way in which detected features will be evaluated

        - as `line` the linewidth is calculated
        - as `circle` an ellipse is fitted and by default features that touch
          the edge of the image get ignored (the default is "circle")

    thresholding : str, optional
        - `auto` uses Otsu's thresholding algorithm
        - `simple` uses a fixed threshold value, 100 by default
          (the default is "auto")

    Returns
    -------
    Pipeline
        the pipeline including default parameters
    """

    if mode == "circle":
        feature_form_filter = FeatureFormFilter(500, 0, True)
        analysis = CircleAnalysis()
    elif mode == "line":
        feature_form_filter = FeatureFormFilter(500, 0, False)
        analysis = LineAnalysis()

    if thresholding == "auto":
        binarization = OtsuThreshold()
    elif thresholding == "simple":
        binarization = BinaryThreshold(100)

    return Pipeline([SimpleReader(),
                     GreyscaleConverter(),
                     GaussianPreprocess(1),
                     binarization,
                     PostprocessNothing(),
                     ContourFinderSimple(),
                     feature_form_filter,
                     analysis])


def make_gui(spim_or_filepath, mode="circle", thresholding="auto"):
    """Creates a :class:`~spotlob.SpotlobNotebookGui` object which opens
    a given :class:`~spotlob.Spim` or image file for preview editing

    PARAMETERS
    ----------
    spim_or_filepath : Spim or str
        Spim (should be cached) or filepath of an image file
        to be loaded

    mode : str
        "circle" or "line" depending on how you want the image to
        be evaluated

    thresholding : str
        "auto" for automatic thresholding based on the histogram (Otsu)
        or "simple" for a fixed thresholding based on a single value

    RETURNS
    -------
    SpotlobNotebookGui
        GUI object that can be displayed using the
        :func:`~spotlob.defaults.show_gui` function within
        a jupyter notebook
    """
    if not isinstance(spim_or_filepath, Spim):
        if os.path.exists(spim_or_filepath):
            spim = Spim.from_file(spim_or_filepath, cached=True)
        else:
            raise FileNotFoundError("File %s not found" % spim_or_filepath)
    else:
        spim = spim_or_filepath

    pipe = default_pipeline(mode=mode, thresholding=thresholding)
    preview_screen = MatplotlibPreviewScreen()
    gui = SpotlobNotebookGui(pipe, preview_screen, spim)
    return gui


def show_gui(gui, *args, **kwargs):
    """Display a :class:`~spotlob.SpotlobNotebookGui` object.
    Run the `%matplotlib widget` magic command to get live preview

    PARAMETERS
    ----------
    gui : SpotlobNotebookGui
        Widget and preview screen, as created by the
        :func:`~spotlob.defaults.show_gui` function
    """
    widgets = gui.make_widgets()
    gui.show_preview_screen(*args, **kwargs)
    display(widgets)
    display(gui.run_button())


def use_in(gui, register=None):
    """Use this as a decorator replace a process in a
    :class:`~spotlob.SpotlobNotebookGui` object

    PARAMETERS
    ----------
    gui : SpotlobNotebookGui
        GUI object to replace the process in

    RETURNS
    -------
    callable
        wrapper function
    """
    if register is None:
        register = PROCESS_REGISTER

    def wrapper(fn):
        process = register.available_processes[fn.__name__]

        # overwrite process at the given stage in pipeline of gui
        gui.pipeline = gui.pipeline.replaced_with(process)
        return fn
    return wrapper


def load_image(filepath, cached=False):
    """Create a :class:`~spotlob.Spim` from a filepath directly,
    using a default reader

    PARAMETERS
    ----------
    filepath : str
        Path to the image file
    cached : bool
        Wether or not references to previous Spim should be kept
        when processing this spim to newer versions

    RETURNS
    -------
    Spim
        Spim containing the image, at stage SpimStage.loaded
    """
    spim = Spim.from_file(filepath, cached=cached)
    reader = SimpleReader()
    return spim.read(reader)



from setuptools import setup
from os.path import join


with open('README.md') as readme_file:
    readme = readme_file.read()

with open(join('spotlob', 'version.py')) as version_file:
    lines = version_file.readlines()
    metadict = dict()

    for line in lines:
        try:
            k, v = line.split('=')
            metadict.update({k.strip(): v.strip().strip('"')})
        except ValueError:
            pass

setup(
    name=metadict['__name__'],
    version=metadict['__version__'],
    author=metadict['__author__'],
    author_email=metadict['__email__'],
    description=metadict['__summary__'],
    license="BSD 3-clause",
    url=metadict['__url__'],
    classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Scientific/Engineering :: Image Recognition',
    ],
    long_description=readme,
    long_description_content_type='text/markdown'
)



import datetime

log_filepath = "process.log"


def write_to_logfile(text):
    with open(log_filepath, "a") as f_:
        f_.write(text)


def log(process_function):
    def wrapper(*args, **kwargs):
        timestamp = datetime.datetime.utcnow()

        entry_str = timestamp.strftime("utc %Y-%m-%d %H:%M:%S ") \
            .join(process_function.im_class.__name__) \
            .join(".") \
            .join(process_function.__name__) \
            .join(str(args[1:])) \
            .join(str(kwargs)) \
            .join("\n")

        write_to_logfile(entry_str)
        return process_function(*args, **kwargs)
    return wrapper



"""
This module provides a widget to browse through the
already applied detection, to see if everything worked
"""
import numpy as np

from ipywidgets import HBox, Button
from IPython.display import display, clear_output

from spotlob.preview import MatplotlibPreviewScreen
from spotlob.spim import Spim, SpimStage


class ReviewBrowser(object):
    index_state = 0
    crop_to_contour = False

    def __init__(self, dataframe, pipeline, crop_to_contour=False):
        self.dataframe = dataframe.reset_index(drop=True)
        self.screen = MatplotlibPreviewScreen()

        stage_dict = pipeline.process_stage_dict

        self.load_process = stage_dict[SpimStage.loaded-1]
        self.analysis_process = stage_dict[SpimStage.analyzed-1]
        self.crop_to_contour = crop_to_contour

    def show_next(self, btn):
        self.index_state = (self.index_state + 1) % len(self.dataframe)
        self.draw()

    def show_prev(self, btn):
        self.index_state = (self.index_state - 1) % len(self.dataframe)
        self.draw()

    def draw(self):
        row = self.dataframe.iloc[self.index_state, :]
        fp = row["filepath"]
        s0 = Spim.from_file(fp, cached=True).read(self.load_process)

        img = np.copy(s0.image)

        results_data = row.to_frame().T
        img_with_res = self.analysis_process.draw_results(img,
                                                          results_data,
                                                          self.crop_to_contour)

        if self.screen.fg_ax is None:
            self.screen.make_new(img_with_res)
        else:
            self.screen.update(img_with_res)

    def drop_current_row(self, btn):
        self.dataframe = self.dataframe.reset_index(drop=True) \
            .drop(self.index_state, axis=0)
        self.index_state = max(self.index_state-1, 0)
        self.draw()


def review_widget(dataframe, pipeline, crop_to_contour=False):
    """Displays a review widget, that has a button
    to browse through all the files listed in a given
    dataframe and draw the result of the analysis contained
    in the given pipeline

    Parameters
    ----------
    dataframe : pandas.Dataframe
        Result from a detection with pipeline
    pipeline : Pipeline
        The pipeline that the results have been created with
    """
    browser = ReviewBrowser(dataframe, pipeline, crop_to_contour)

    next_button = Button(description="next")
    next_button.on_click(browser.show_next)

    drop_button = Button(description="drop")
    drop_button.on_click(browser.drop_current_row)

    prev_button = Button(description="prev")
    prev_button.on_click(browser.show_prev)

    display(HBox([prev_button, drop_button, next_button]))
    return browser
